/*
 * Created on Oct 08, 2004
 * Created by Alon Rohter
 * Copyright (C) Azureus Software, Inc, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package com.biglybt.core.util.average;

/**
 * Implements a simple running average.
 */
public class RunningAverage implements Average {

   private long count = 0;
   private double sum;

   /**
    * Create a new running average.
    */
   public RunningAverage() {
      this.sum = 0;
   }

	@Override
	public void reset(){
		count = 0;
		sum = 0;
	}
   /**
    * Update average and return average-so-far.
    */
   @Override
   public double update(final double newValue) {
      sum += newValue;
      count++;
      return sum / count;
   }

   /**
    * Return average-so-far.
    */
   @Override
   public double getAverage() { return count==0?0:(sum / count ); }

}
