/*
 * Created on 14-Jun-2005
 * Created by Paul Gardner
 * Copyright (C) Azureus Software, Inc, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

package com.biglybt.plugin.dht.impl;

import com.biglybt.core.dht.transport.DHTTransportValue;
import com.biglybt.plugin.dht.DHTPluginValue;


public class
DHTPluginValueImpl
	implements DHTPluginValue
{
	private DHTTransportValue		value;

	protected
	DHTPluginValueImpl(
		DHTTransportValue	_value )
	{
		value	= _value;
	}

	@Override
	public byte[]
	getValue()
	{
		return( value.getValue());
	}

	@Override
	public long
	getCreationTime()
	{
		return( value.getCreationTime());
	}

	@Override
	public long
	getVersion()
	{
		return( value.getVersion());
	}

	@Override
	public boolean
	isLocal()
	{
		return( value.isLocal());
	}

	@Override
	public int
	getFlags()
	{
		return( value.getFlags()&0xff);
	}
}