/*
 * File    : TrackerTorrentRequestImpl.java
 * Created : 14-Dec-2003
 * By      : parg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details ( see the LICENSE file ).
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.biglybt.pifimpl.local.tracker;

/**
 * @author parg
 *
 */

import java.util.Map;

import com.biglybt.core.tracker.host.TRHostPeer;
import com.biglybt.core.tracker.host.TRHostTorrent;
import com.biglybt.core.tracker.host.TRHostTorrentRequest;
import com.biglybt.pif.tracker.TrackerPeer;
import com.biglybt.pif.tracker.TrackerTorrent;
import com.biglybt.pif.tracker.TrackerTorrentRequest;

public class
TrackerTorrentRequestImpl
	implements TrackerTorrentRequest
{
	protected TRHostTorrentRequest		req;

	protected
	TrackerTorrentRequestImpl(
		TRHostTorrentRequest 	_req )
	{
		req	= _req;
	}

	@Override
	public int
	getRequestType()
	{
		if ( req.getRequestType() == TRHostTorrentRequest.RT_ANNOUNCE ){

			return( RT_ANNOUNCE );

		}else if ( req.getRequestType() == TRHostTorrentRequest.RT_SCRAPE ){

			return( RT_SCRAPE );

		}else{

			return( RT_FULL_SCRAPE );
		}
	}

	@Override
	public TrackerTorrent
	getTorrent()
	{
		TRHostTorrent	torrent = req.getTorrent();

		if ( torrent == null ){

			return( null );
		}

		return( new TrackerTorrentImpl( torrent ));
	}

	@Override
	public TrackerPeer
	getPeer()
	{
		TRHostPeer	peer = req.getPeer();

		if ( peer == null ){

			return( null );
		}

		return( new TrackerPeerImpl( peer ));
	}

	@Override
	public String
	getRequest()
	{
		return( req.getRequest());
	}

	@Override
	public Map
	getResponse()
	{
		return( req.getResponse());
	}
}
